.packageName <- "externalVector"
library('methods')

## These two functions would be replaced by "attributes"
## and "attributes<-" if and when slots are no longer
## stored as attributes

nonSlotAttributes <-
    function(x)
{
    ans <- attributes(x)
    klass <- class(x)
    if (isClass(klass)) {
        ans[c(slotNames(klass), "class")] <- NULL
    }
    ans
}

"nonSlotAttributes<-" <-
    function(x, value)
{
    if (!(is.null(value) || is.list(value) || is.pairlist(value)))
        stop("attributes must be in a list")
    klass <- class(x)
    if (isClass(klass)) {
        if (length(value) == 0) {
            value <- attributes(x)[c(slotNames(klass), "class")]
        } else {
            nm <- c(slotNames(klass), "class")
            value[nm] <- attributes(x)[nm]
        }
    }
    attributes(x) <- value
    x
}

if (length(findFunction("setVirtualMethod")) == 0) {
    setVirtualMethod <-
        function (name, signature.virtual,
                  signature.nonvirtual = character(),
                  ..., where=topenv(parent.frame()))
    {
        if (length(signature.virtual) == 0)
            stop("virtual method signature must be of non-zero length")
        if (!any(unlist(lapply(signature.virtual, isVirtualClass))))
            stop("virtual method signature can only have VIRTUAL classes")

        f <- getGeneric(name)
        if (is.null(f)) {
            setGeneric(name, ..., where=where)
            f <- getGeneric(name)
        }

        if (length(names(signature.virtual)) !=
            length(signature.virtual) && length(signature.nonvirtual) != 0)
            stop("can not combine virtual and non-virtual parts. ",
                 "Provide names for virtual arguments")
        if (length(names(signature.nonvirtual)) !=
            length(signature.nonvirtual))
            stop("can not combine virtual and non-virtual parts. ",
                 "Provide names for non-virtual arguments")
        signature.method <- 
            c(signature.virtual, signature.nonvirtual)
        sig <- f@signature
        attributes(f) <- NULL
             
        ## Find the virtual parameter names
        mcall <-
            as.list(seq(along=signature.method))
        names(mcall) <- names(signature.method)
        mcall <- do.call("call", c(name, mcall))
        mcall <- match.call(f, mcall)
        mcall <- as.list(mcall[-1])
        nm <- names(mcall)
        notmatched <- !(nm %in% sig)
        if (any(notmatched))
            stop("unused arguments in signature: ", nm[notmatched])
        mcall <- unlist(mcall)
        names(signature.method) <- names(sort(mcall))
        signature.method <- as.list(signature.method)[sig]
        signature.method[unlist(lapply(signature.method,
                                       is.null))] <- NULL
        signature.method <-
            do.call("signature", signature.method)

        ## Find the virtual signature names
        parNames <- names(sort(mcall[mcall <= length(signature.virtual)]))

        ## Create the method definition
        parname <- as.name(paste('.', sort(parNames)[1], sep=''))
        parDef <-
            bquote(function(x) paste(.(parname), ' = "', class(get(.(parname))), '"', sep = ''))
        names(parDef[[2]]) <- as.character(parname)

        body(f) <- bquote({
            stop("Trying to call virtual method for \"", .(name),
                 "\" with signature ",
                 paste(lapply(.(parNames), .(parDef)), collapse = ', '))
        })
        setMethod(name, signature.method, f, where=where)
    }
}

setClass("externalVector", "VIRTUAL")

setClassUnion("vectorNamesType", c("character", "NULL"))
setClassUnion("integerOrNull", c("integer", "NULL"))
setClassUnion("listOrNull", c("list", "NULL"))

if (!isGeneric("any.na")) {
    any.na <- function(x) any(is.na(x))
    setGeneric("any.na")
}

if (!isGeneric("all.na")) {
    all.na <- function(x) all(is.na(x))
    setGeneric("all.na")
}
##
## Memory allocators for external resources.
##

setClass("externalAllocator")

setClass("gcAllocator",
         contains="externalAllocator")

##
## Base class for external resource allocated by the allocators
##

setClass("externalResource")

setMethod("initialize",
          signature(.Object = "externalResource"),
          function (.Object, ...)
          allocate(.Object, ...))

##
## Return a suitable allocator for the resource
##
if (!isGeneric("allocator"))
    setGeneric("allocator",
               function(resource)
               standardGeneric("allocator"))

##
## Each resource type should define its default allocator
##
setVirtualMethod("allocator",
          signature(resource="externalResource"))

if (!isGeneric("allocatedType"))
    setGeneric("allocatedType",
               function(resource)
               standardGeneric("allocatedType"))

setMethod("allocatedType",
          signature("externalResource"),
          function (resource) NULL)

if (!isGeneric("allocatedSize"))
    setGeneric("allocatedSize",
               function(resource)
               standardGeneric("allocatedSize"))

setVirtualMethod("allocatedSize",
                 signature(resource="externalResource"))

##
## Allocate some raw memory for the resource using allocator alloc,
## create an external pointer with that raw memory, initialize the
## resource with the pointer and return it. If type is a vector type
## (one of logical, integer, numeric, complex, character or list) then
## size is the length of the vector. Otherwise, size is in bytes.
##
## The allocated memory is not initialized.
##
if (!isGeneric("allocate"))
    setGeneric("allocate",
               function(resource, alloc, size, type, ...)
               standardGeneric("allocate"))

setVirtualMethod("allocate",
                 signature(alloc="externalAllocator"),
                 signature(resource="externalResource"))

setMethod("allocate",
          signature(resource="externalResource", alloc="missing"),
          function (resource, alloc, size, type, ...)
      {
          alloc <- allocator(resource)
          if (missing(size))
              size <- allocatedSize(resource)
          if (missing(type))
              type <- allocatedType(resource)
          allocate(resource, alloc=alloc, size=size, type=type, ...)
      })

##
## This can be called explicitly to (potentially) free the memory
## 
if (!isGeneric("deallocate"))
    setGeneric("deallocate",
               function(resource, alloc)
               standardGeneric("deallocate"))

setVirtualMethod("deallocate",
                 signature(alloc="externalAllocator"),
                 signature(resource="externalResource"))

setMethod("deallocate",
          signature(resource="externalResource",
                    alloc="missing"),
          function (resource, alloc)
      {
          deallocate(resource, alloc=allocator(resource))
      })

##
## Given a resource and an external pointer allocated for that
## resource, initialize and return the resource.
##
if (!isGeneric("initializeResource"))
    setGeneric("initializeResource",
               function(resource, ptr, size, type, ...)
               standardGeneric("initializeResource"))

setVirtualMethod("initializeResource",
                 signature(resource="externalResource"),
                 signature(ptr="externalptr"))

##
## Retrieve the external pointer associated with the resource
##
if (!isGeneric("getPointer"))
    setGeneric("getPointer",
               function(resource)
               standardGeneric("getPointer"))

setVirtualMethod("getPointer",
                 signature(resource="externalResource"))

if (!isGeneric("reinitializePointer"))
    setGeneric("reinitializePointer",
               function(resource, alloc)
               standardGeneric("reinitializePointer"))

setVirtualMethod("reinitializePointer",
                 signature(alloc="externalAllocator"),
                 signature(resource="externalResource"))

setMethod("reinitializePointer",
          signature(resource="externalResource",
                    alloc="missing"),
          function (resource, alloc) 
          reinitializePointer(resource, alloc=allocator(resource)))

if (!isGeneric("external.size"))
    setGeneric("external.size",
               function(resource, alloc)
               standardGeneric("external.size"))

setVirtualMethod("external.size",
                 signature(alloc="externalAllocator"),
                 signature(resource="externalResource"))

setMethod("external.size",
          signature(resource="externalResource",
                    alloc="missing"),
          function (resource, alloc) 
          external.size(resource, alloc=allocator(resource)))

if (!isGeneric("external.size<-"))
    setGeneric("external.size<-",
               function(resource,
                        copy = TRUE, alloc,
                        value)
               standardGeneric("external.size<-"))

setVirtualMethod("external.size<-",
                 signature(alloc="externalAllocator"),
                 signature(resource="externalResource",
                           copy="logical"))

setReplaceMethod("external.size",
                 signature(resource = "externalResource",
                           alloc = "missing"),
                 function (resource, copy, alloc, value) 
             {
                 if (missing(copy))
                     copy <- TRUE
                 "external.size<-"(resource, copy,
                                   allocator(resource), value)
             })

##
## Methods for gcAllocator
##
setMethod("allocate",
          signature(resource="externalResource",
                    alloc="gcAllocator"),
          function(resource, alloc, size, type, ...)
      {
          ptr <- .Call("gcExternalPtr_new", size, type,
                       PACKAGE="externalVector")
          initializeResource(resource, ptr, size, type, ...)
      })

setMethod("reinitializePointer",
          signature(resource="externalResource",
                    alloc="gcAllocator"),
          function (resource, alloc) 
      {
          .Call("gcExternalPtr_restore", getPointer(resource),
                PACKAGE="externalVector")
          resource
      })

setMethod("external.size",
          signature(resource="externalResource",
                    alloc="gcAllocator"),
          function(resource, alloc)
      {
          .Call("gcExternalPtr_size", getPointer(x),
                PACKAGE="externalVector")
      })

setReplaceMethod("external.size",
                 signature(resource="externalResource",
                           copy = "logical",
                           alloc = "gcAllocator",
                           value="numeric"),
                 function(resource, copy, alloc, value)
             {
                 .Call("gcExternalPtr_resize", getPointer(resource),
                       value, copy,
                       PACKAGE="externalVector")
                 resource
             })

setMethod("deallocate",
          signature(resource="externalResource",
                    alloc="gcAllocator"),
          function(resource, alloc)
      {
          .Call("gcExternalPtr_delete", getPointer(resource))
          invisible(NULL)
      })
setClass("vectorIndex")

setVirtualMethod("Arith",
                 signature(e1 = "vectorIndex"))

setVirtualMethod("Arith",
                 signature(e2 = "vectorIndex"))

setClass("NAIndex",
         representation(Length="integer", Names="vectorNamesType"),
         contains = c("vectorIndex", "logical"),
         prototype=structure(NA, length=as.integer(1),
                             Names = NULL))

setMethod("initialize",
          signature(.Object = "NAIndex"),
          function (.Object, Length = 0, ...)
      {
          .Object <- .Call("initNAIndex", Length,
                           PACKAGE="externalVector")
          if (!missing(Names)) {
              Names <- asEach(Names, "character")
              if (length(.Object) < length(Names))
                  stop("Names slot must be the same length as the vector")
              else if (length(.Object) > length(Names))
                  length(Names) <- length(.Object)
              .Object@Names <- Names
          }
          .Object
      })

setClass("allIndex",
         contains=c("vectorIndex", "logical"),
         prototype=TRUE)

setMethod("initialize",
          signature(.Object = "allIndex"),
          function (.Object, ...) .Call("initAllIndex",
                                        PACKAGE="externalVector"))

setClass("noneIndex", contains=c("vectorIndex", "integer"),
         prototype=integer(0))

setMethod("initialize",
          signature(.Object = "noneIndex"),
          function (.Object, ...) .Call("initNoneIndex",
                                        PACKAGE="externalVector"))

setClass("sliceIndex",
         representation(content="numeric", Names="vectorNamesType"),
         contains="vectorIndex",
         prototype=list(content=as.numeric(rep(1, 3))))

setClass("positiveIndex",
         representation(rangeIndex="integer",
                        noNA="logical"),
         contains = c("vectorIndex", "integer"),
         prototype = structure(integer(0),
                               rangeIndex=as.integer(c(NA, NA)),
                               noNA=NA))

setMethod("initialize",
          signature(.Object = "positiveIndex"),
          function (.Object, ...) {
              .External('initPositiveIndex', ...,
                    PACKAGE='externalVector')
          })

setMethod("names",
          signature(x="NAIndex"),
          function (x) x@Names)

setMethod("names",
          signature(x="sliceIndex"),
          function (x) x@Names)

setReplaceMethod("names",
                 signature(x = "NAIndex"),
                 function (x, value) 
             {
                 x@Names <- as.character(value)
                 x
             })

setReplaceMethod("names",
                 signature(x = "NAIndex"),
                 function (x, value) 
             {
                 if (!is.character(value))
                     value <- as.character(value)
                 if (length(value) > length(x))
                     stop("Names slot must be the same length as the externalVector")
                 if (length(value) < length(x))
                     length(value) <- length(x)
                 x@Names <- value
                 x
             })

setReplaceMethod("names",
                 signature(x = "NAIndex", value = "externalVector"),
                 function (x, value) 
             {
                 if (class(defaultElement(value)) != "character")
                     value <- asEach(value, "character")
                 if (length(value) > length(x))
                     stop("Names slot must be the same length as the externalVector")
                 if (length(value) < length(x))
                     length(value) <- length(x)
                 x@Names <- x
                 x
             })

setReplaceMethod("names",
                 signature(x = "sliceIndex"),
                 function (x, value) 
             {
                 x@Names <- as.character(value)
                 x
             })

setReplaceMethod("names",
                 signature(x = "sliceIndex"),
                 function (x, value) 
             {
                 if (!is.character(value))
                     value <- as.character(value)
                 if (length(value) > length(x))
                     stop("Names slot must be the same length as the externalVector")
                 if (length(value) < length(x))
                     length(value) <- length(x)
                 x@Names <- value
                 x
             })

setReplaceMethod("names",
                 signature(x = "sliceIndex", value = "externalVector"),
                 function (x, value) 
             {
                 if (class(defaultElement(value)) != "character")
                     value <- asEach(value, "character")
                 if (length(value) > length(x))
                     stop("Names slot must be the same length as the externalVector")
                 if (length(value) < length(x))
                     length(value) <- length(x)
                 x@Names <- x
                 x
             })

setMethod("any.na",
          signature(x="allIndex"),
          function (x) FALSE)

setMethod("any.na",
          signature(x="noneIndex"),
          function (x) FALSE)

setMethod("any.na",
          signature(x="NAIndex"),
          function (x) TRUE)

setMethod("any.na",
          signature(x="sliceIndex"),
          function (x) FALSE)

setMethod("any.na",
          signature(x="positiveIndex"),
          function (x)
          if (is.na(x@noNA)) {
              any(is.na(x))
          } else !x@noNA)

setMethod("all.na",
          signature(x="allIndex"),
          function (x) FALSE)

setMethod("all.na",
          signature(x="noneIndex"),
          function (x) FALSE)

setMethod("all.na",
          signature(x="NAIndex"),
          function (x) TRUE)

setMethod("all.na",
          signature(x="sliceIndex"),
          function (x) FALSE)

setMethod("all.na",
          signature(x="positiveIndex"),
          function (x)
          (is.na(x@noNA) || !x@noNA) && all(is.na(x)))

setMethod("length",
          signature(x = "NAIndex"),
          function (x) x@Length)

setReplaceMethod("length",
                 signature(x = "NAIndex"),
                 function (x, value) 
             {
                 n <- as.integer(value)
                 if (is.na(value))
                     stop("missing value for length")
                 if (length(value) > 1)
                     stop("invalid second argument")
                 if (length(value) != 1)
                     stop("invalid second argument")
                 if (value < 0)
                     stop("negative length vectors are not allowed")
                 if (value == 0)
                     new("noneIndex")
                 else {
                     x@Length <- n
                     if (!is.null(names(x)))
                         length(x@Names) <- n
                     x
                 }
             })

setMethod("[",
          signature(x = "NAIndex"),
          function (x, i, j, ..., drop) 
      {
          length(x) <- length(i)
          x
      })

setMethod("[",
          signature(x = "NAIndex", i = "noneIndex"),
          function (x, i, j, ..., drop) 
      {
          i
      })

setMethod("[",
          signature(x = "NAIndex", i = "allIndex"),
          function (x, i, j, ..., drop) 
      {
          x
      })

setReplaceMethod("[",
                 signature(x = "NAIndex"),
                 function (x, i, j, ..., value) 
             {
                 stop("object is not subassignable")
             })

setReplaceMethod("[[",
                 signature(x = "NAIndex"),
                 function (x, i, j, ..., value) 
             {
                 stop("object is not subassignable")
             })

setMethod("length",
          signature(x = "allIndex"),
          function (x) 
      {
          as.integer(Inf)
      })

setReplaceMethod("length",
                 signature(x = "allIndex"),
                 function (x, value) 
             {
                 stop("can not change length of object")
             })

setMethod("[",
          signature(x = "allIndex", i = "ANY",
                    j = "missing", drop = "missing"),
          function (x, i, j, ..., drop) 
      {
          i
      })

setMethod("[",
          signature(x = "allIndex", i = "missing",
                    j = "missing", drop = "missing"),
          function (x, i, j, ..., drop) 
      {
          x
      })

setReplaceMethod("[",
                 signature(x = "allIndex"),
                 function (x, i, j, ..., value) 
             {
                 stop("object is not subassignable")
             })

setReplaceMethod("[[",
                 signature(x = "allIndex"),
                 function (x, i, j, ..., value) 
             {
                 stop("object is not subassignable")
             })

setMethod("length",
          signature("noneIndex"),
          function (x) 
      {
          as.integer(0)
      })

setMethod("[",
          signature(x = "noneIndex", i = "missing",
                    j = "missing", drop = "missing"),
          function (x, i, j, ..., drop) 
      {
          x
      })

setReplaceMethod("length",
                 signature(x = "noneIndex"),
                 function (x, value) 
             {
                 stop("can not change length of object")
             })

setReplaceMethod("[",
                 signature(x = "noneIndex"),
                 function (x, i, j, ..., value) 
             {
                 stop("object is not subassignable")
             })

setReplaceMethod("[[",
                 signature(x = "noneIndex"),
                 function (x, i, j, ..., value) 
             {
                 stop("object is not subassignable")
             })

setMethod("[",
          signature(x = "noneIndex", i = "ANY",
                    j = "missing", drop = "missing"),
          function (x, i, j, ..., drop) 
      {
          new("NAIndex", length=length(i))
      })

setMethod("[",
          signature(x = "noneIndex", i = "logical",
                    j = "missing", drop = "missing"),
          function (x, i, j, ..., drop) 
      {
          new("NAIndex", length=sum(is.na(i), i, na.rm=TRUE))
      })

setMethod("length",
          signature(x = "sliceIndex"),
          function (x) 
      {
          as.integer(x@content[2])
      })

# setMethod("rev",
#           signature(x="sliceIndex"),
#           function (x) 
#       {
#           if (length(x) > 1) {
#               e1@content[3] = -e1@content[3]
#               e1@content[1] = e1@content[1]-e1@content[3]*(e1@content[2]-1)
#           }
#           x
#       })

setReplaceMethod("length",
                 signature(x="sliceIndex"),
                 function(x, value)
             {
                 if (is.nan(value) || value == -Inf)
                     stop('length value invalid')
                 if (value < 1.0)
                     return(new("noneIndex"))
                 value <- trunc(as.numeric(value))
                 if (value != x@content[1]) {
                     x@content[1] <- value
                     if (!is.null(names(x)))
                         length(x@Names) <- value
                 }
                 x
             })

setMethod("[",
          signature(x = "sliceIndex", i = "allIndex",
                    j = "missing"),
          function (x, i, j, ..., drop) 
      {
          x
      })

setMethod("[",
          signature(x = "sliceIndex", i = "noneIndex",
                    j = "missing"),
          function (x, i, j, ..., drop) 
      {
          i
      })

setMethod("[",
          signature(x = "sliceIndex", i = "sliceIndex",
                    j = "missing"),
          function (x, i, j, ..., drop) 
      {
          contx <- x@content
          conti <- i@content
          start <- conti[1]
          len <- conti[2]
          end <- conti[1]+conti[3]*(len-1)
          if (start >= 0 && end < contx[2]) {
              conti[1] <- contx[1]+start*contx[3]
              conti[3] <- contx[3]*conti[3]
              x@content <- conti
              if (!is.null(names(x)))
                  names(x) <- Subset(names(x), i)
              x
          } else as(x, "positiveIndex")[i]
      })

setMethod("[",
          signature(x = "sliceIndex", i = "positiveIndex",
                    j = "missing"),
          function (x, i, j, ..., drop) 
      {
          nm <- x@Names
          contx <- x@content
          start <- contx[1]
          len <- contx[2]
          stride <- contx[3]

          noNA <- i@noNA
          if (is.na(noNA)) {
              na <- is.na(i)
              noNA <- any(na)
              if (noNA)
                  na <- logical(0)
          } else if (noNA)
              na <- logical(0)
          else na <- is.na(i)

          range.i <- i@rangeIndex
          if (any.na(range.i))
              range.i <- as.integer(range(i, na.rm=T))

          if (range.i[1] < 0 || range.i[2] >= len) {
              if (noNA) {
                  na <- i < 0 | i >= len
                  if (!any(na))
                      na <- logical(0)
              } else na[!na] <- i[!na] < 0 | i[!na] >= len
          }
          if (length(na) == 0) {
              ans <- start+i*stride
          } else {
              ans <- start+i*stride
              ans[na] <- as.integer(NA)
              ans@noNA <- FALSE
          }
          if (!is.null(nm))
              names(ans) <- Subset(nm, i)
          ans
      })

setMethod("[",
          signature(x = "sliceIndex",
                    j = "missing"),
          function (x, i, j, ..., drop) 
      {
          as(x, "positiveIndex")[i]
      })

setMethod("[",
          signature(x = "positiveIndex", i = "sliceIndex",
                    j = "missing"),
          function (x, i, j, ..., drop) 
      {
          conti <- as.integer(i@content)
          start <- conti[1]
          len <- conti[2]
          stride <- conti[3]
          if (len > 0) {
              i <- start+1+(0:(len-1))*stride
              as(as.integer(x)[i], "positiveIndex")
          } else new("positiveIndex", integer(0))
      })

setMethod("[",
          signature(x = "positiveIndex", i = "positiveIndex",
                    j = "missing"),
          function (x, i, j, ..., drop) 
      {
          nm <- names(x)
          if (!is.null(nm))
              nm <- Subset(nm, i)
          i <- unclass(i+1)
          x <- new("positiveIndex", unclass(x)[i])
          names(x) <- nm
          x
      })

setMethod("[",
          signature(x = "positiveIndex",
                    j = "missing"),
          function (x, i, j, ..., drop) 
      {
          as(Subset(as.integer(x), i), "positiveIndex")
      })

makeSlice <-
    function(start=1, length.slice=1, stride=1)
{
    .Call("makeSlice", start, length.slice, stride,
          PACKAGE="externalVector")
}

setAs("integer", "positiveIndex",
      function(from, to) {
          if (any(from < 0, na.rm=TRUE))
              stop("negative integers can not be coerced to positiveIndex")
          from <- from[from != 0]
          nm <- names(from)
          names(from) <- NULL
          ans <- new("positiveIndex", as.integer(from-1))
          names(ans) <- nm
          ans
      })

if (!isGeneric("as.integer"))
    setGeneric("as.integer")

setMethod("as.integer",
          signature(x="positiveIndex"),
          function (x, ...) 
      {
          nm <-names(x)
          x <- as.integer(as.vector(x)+1)
          names(x) <- nm
          x
      })

setAs("numeric", "positiveIndex",
      function(from, to) {
          as(as.integer(from), "positiveIndex")
      })

setAs("sliceIndex", "positiveIndex",
      function(from, to) {
          cont <- as.integer(from@content)
          n <- cont[2]
          ans <- if (n > 0)
              new("positiveIndex",
                  cont[1]+(0:(n-1))*cont[3])
          else new("positiveIndex", integer(0))
          names(ans) <- names(from)
          ans
      })

setMethod("Arith",
          signature(e1 = "noneIndex", e2 = "numeric"),
          function (e1, e2) e1)

setMethod("Arith",
          signature(e1 = "numeric", e2 = "noneIndex"),
          function (e1, e2) e2)

setMethod("Arith",
          signature(e1 = "NAIndex", e2 = "numeric"),
          function (e1, e2) e1)

setMethod("Arith",
          signature(e1 = "numeric", e2 = "NAIndex"),
          function (e1, e2) e2)

setMethod("Arith",
          signature(e1 = "allIndex", e2 = "numeric"),
          function (e1, e2) e1)

setMethod("Arith",
          signature(e1 = "numeric", e2 = "allIndex"),
          function (e1, e2) e2)

setMethod("Arith",
          signature(e1 = "sliceIndex", e2 = "numeric"),
          function (e1, e2) 
      {
          if (length(e2) != 1 || !(.Generic %in% c('+', '*')))
              stop('unimplemented operation')
          if (is.na(as.integer(e2))) {
              new("NAIndex", length=as.integer(e1@content[2]),
                  Names = names(e1))
          } else {
              if (.Generic == '+')
                  e1@content[1] <- e1@content[1]+e2
              else e1@content[c(1, 3)] <- e1@content[c(1, 3)]*e2
              if (e1@content[1] < 0 ||
                  e1@content[1]+(e1@content[2]-1)*e1@content[3] < 0)
                  stop("negative index not allowed")
              e1
          }
      })

setMethod("Arith",
          signature(e1 = "numeric", e2 = "sliceIndex"),
          function (e1, e2) 
          callGeneric(e2, e1))
          
setMethod("Arith",
          signature(e1 = "positiveIndex", e2 = "numeric"),
          function (e1, e2)
      {
          if (length(e2) != 1 || !(.Generic %in% c('+', '*')))
              stop('unimplemented operation')
          if (is.na(as.integer(e2)))
              new("NAIndex", length=length(e1), Names = names(e1))
          else new("positiveIndex",
                   callNextMethod(e1, as.integer(e2)))
      })

setMethod("Arith",
          signature(e1 = "numeric", e2 = "positiveIndex"),
          function (e1, e2) 
          callGeneric(e2, e1))
          
SubscriptList <-
    function(subs, length.object, dim.object=NULL,
             names.object=NULL, dimnames.object=NULL,
             assign.operation=FALSE)
{
    .Call("SubscriptList", subs, length.object, dim.object,
          names.object, dimnames.object, assign.operation,
          PACKAGE='externalVector')
}

if (!isGeneric("Subset"))
    setGeneric("Subset",
               function(x, i) standardGeneric("Subset"))

setMethod("Subset",
          signature("ANY"),
          function (x, i) 
      {
          x[i]
      })

setMethod("Subset",
          signature(x = "ANY", i = "allIndex"),
          function (x, i) x)

setMethod("Subset",
          signature(x = "ANY", i = "noneIndex"),
          function (x, i) x[NULL])

setMethod("Subset",
          signature(x = "ANY", i = "NAIndex"),
          function (x, i)
      {
          na <- x[length(x)+1]
          length(x) <- length(i)
          x[] <- na
          x
      })

setMethod("Subset",
          signature(x = "ANY", i = "positiveIndex"),
          function (x, i)
      {
          zero <- i == 0
          i[zero] <- NA
          i <- x[-1][i]
          i[zero] <- x[1]
          i
      })

setMethod("Subset",
          signature(x = "ANY", i = "sliceIndex"),
          function (x, i)
      {
          cont <- as.integer(i@content)
          start <- cont[1]
          n <- cont[2]
          stride <- cont[3]
          x[start+1+(0:(n-1))*stride]
      })

if (!isGeneric("Subset2"))
    setGeneric("Subset2",
               function(x, i, j) standardGeneric("Subset2"))

setMethod("Subset2",
          signature("ANY"),
          function (x, i, j) 
      {
          x[i, j, drop = FALSE]
      })

setMethod("Subset2",
          signature(x = "ANY", i = "allIndex", j = "allIndex"),
          function (x, i, j) x)

setMethod("Subset2",
          signature(x = "ANY", i = "allIndex"),
          function (x, i, j) x[, j, drop=FALSE])

setMethod("Subset2",
          signature(x = "ANY", j = "allIndex"),
          function (x, i, j) x[i, , drop=FALSE])

setMethod("Subset2",
          signature(x = "ANY", i = "noneIndex"),
          function (x, i, j) Subset(x, NULL, j))

setMethod("Subset2",
          signature(x = "ANY", j = "noneIndex"),
          function (x, i, j) Subset(x, i, NULL))

setMethod("Subset2",
          signature(x = "ANY", i = "noneIndex", j = "noneIndex"),
          function (x, i, j) x[NULL, NULL, drop = FALSE])

setMethod("Subset2",
          signature(x = "ANY", i = "NAIndex", j = "NAIndex"),
          function (x, i, j)
      {
          na <- x[length(x)+1]
          length(x) <- length(i)*length(j)
          dim(x) <- c(length(i), length(j))
          x[,] <- na
          x
      })

setMethod("Subset2",
          signature(x = "ANY", i = "NAIndex"),
          function (x, i, j)
      {
          na <- x[length(x)+1]
          length(x) <- length(i)*length(j)
          dim(x) <- c(length(i), length(j))
          x[,] <- na
          x
      })

setMethod("Subset2",
          signature(x = "ANY", j = "NAIndex"),
          function (x, i, j)
      {
          na <- x[length(x)+1]
          length(x) <- length(i)*length(j)
          dim(x) <- c(length(i), length(j))
          x[,] <- na
          x
      })

setMethod("Subset2",
          signature(x = "ANY", i = "NAIndex", j = "allIndex"),
          function (x, i, j)
      {
          na <- x[length(x)+1]
          n <- ncol(x)
          length(x) <- length(i)*n
          dim(x) <- c(length(i), n)
          x[,] <- na
          x
      })

setMethod("Subset2",
          signature(x = "ANY", i = "allIndex", j = "NAIndex"),
          function (x, i, j)
      {
          na <- x[length(x)+1]
          m <- nrow(x)
          length(x) <- m*length(j)
          dim(x) <- c(m, length(j))
          x[,] <- na
          x
      })

setMethod("Subset2",
          signature(x = "ANY", i = "positiveIndex", j = "positiveIndex"),
          function (x, i, j)
      {
          x[unclass(i+1), unclass(j+1), drop = FALSE]
      })

setMethod("Subset2",
          signature(x = "ANY", i = "positiveIndex"),
          function (x, i, j)
      {
          Subset2(x, unclass(i+1), j)
      })

setMethod("Subset2",
          signature(x = "ANY", j = "positiveIndex"),
          function (x, i, j)
      {
          Subset2(x, i, unclass(j+1))
      })

setMethod("Subset2",
          signature(x = "ANY", i = "sliceIndex", j = "sliceIndex"),
          function (x, i, j)
      {
          cont <- as.integer(i@content)
          start1 <- cont[1]
          n1 <- cont[2]
          stride1 <- cont[3]
          cont <- as.integer(j@content)
          start2 <- cont[1]
          n2 <- cont[2]
          stride2 <- cont[3]
          x[start1+1+(0:(n1-1))*stride1,
            start2+1+(0:(n2-1))*stride2, drop=FALSE]
      })

setMethod("Subset2",
          signature(x = "ANY", i = "sliceIndex"),
          function (x, i, j)
      {
          cont <- as.integer(i@content)
          start <- cont[1]
          n <- cont[2]
          stride <- cont[3]
          Subset2(x, start+1+(0:(n-1))*stride, j)
      })

setMethod("Subset2",
          signature(x = "ANY", j = "sliceIndex"),
          function (x, i, j)
      {
          cont <- as.integer(j@content)
          start <- cont[1]
          n <- cont[2]
          stride <- cont[3]
          Subset2(x, i, start+1+(0:(n-1))*stride)
      })
setClass("nativeStorageMethods",
         representation(ptr="externalptr"),
         contains = "externalResource")

setMethod("initializeResource",
          signature(resource="nativeStorageMethods"),
          function (resource, ptr, size, type, ...) 
      {
          resource@ptr <- ptr
          resource
      })

setMethod("allocator",
          signature(resource="nativeStorageMethods"),
          function (resource) new("gcAllocator"))

setMethod("getPointer",
          signature(resource="nativeStorageMethods"),
          function (resource) resource@ptr)

setMethod("allocatedSize",
          signature(resource="nativeStorageMethods"),
          function (resource)
          .Call("storageMethodsStructSize", PACKAGE="externalVector"))

setClass("nativeStorageMethodsList",
         representation(allMethods="nativeStorageMethods",
                        numericMethods="nativeStorageMethods",
                        integerMethods="nativeStorageMethods",
                        logicalMethods="nativeStorageMethods",
                        complexMethods="nativeStorageMethods",
                        characterMethods="nativeStorageMethods",
                        listMethods="nativeStorageMethods"))

setMethod("initialize",
          signature(.Object = "nativeStorageMethodsList"),
          function (.Object, ...) 
      {
          .Object@allMethods <- new("nativeStorageMethods")
          .Object@logicalMethods <- new("nativeStorageMethods")
          .Object@integerMethods <- new("nativeStorageMethods")
          .Object@numericMethods <- new("nativeStorageMethods")
          .Object@complexMethods <- new("nativeStorageMethods")
          .Object@characterMethods <- new("nativeStorageMethods")
          .Object@listMethods <- new("nativeStorageMethods")
          .Object
      })

setClass("externalStorage",
         representation(ptr="externalptr",
                        type="vector",
                        length="integer"),
         contains = c("VIRTUAL", "externalResource"),
         prototype = list(type = logical(1),
                          length = as.integer(0)))

setMethod("allocator",
          signature(resource="externalStorage"),
          function (resource) 
          new("gcAllocator"))

setMethod("initialize",
          signature(.Object="externalStorage"),
          function(.Object,
                   type = logical(1),
                   length = as.integer(0), ...)
      {
          if (!(class(type) %in% c("logical", "integer", "numeric",
                                   "complex", "character",
                                   "list")))
              stop("invalid storage type with class ", class(type))
          if (length(type) < 1)
              .Object@type <- vector(class(type), 1)
          else if (length(type) > 1)
              .Object@type <- type[1]
          else .Object@type <- type
          if (length < 1)
              .Object@length <- as.integer(0)
          else .Object@length <- as.integer(length)
              
          callNextMethod()
      })

setGeneric("getNativeStorageMethods",
           function(x) standardGeneric("getNativeStorageMethods"))

setExternalStorageClass <-
local({
    .nativeMethods <- new.env()

    assign("externalStorage",
           delay(.Call("initNativeStorageMethods",
                       new("nativeStorageMethodsList"),
                       PACKAGE="externalVector")),
           envir = .nativeMethods)

    setMethod("getNativeStorageMethods", "externalStorage",
              function(x)
          {
              Class <- class(x)
              if (exists(Class, envir=.nativeMethods, inherits=FALSE)) {
                  methods <- get(class(x), envir=.nativeMethods)
                  getPointer(slot(methods,
                                  paste(class(internalType(x)),
                                        "Methods", sep="")))
              } else {
                  stop("could not find storage methods for class \"",
                       Class, "\"")
              }
          })

    setNativeStorageMethods <-
        function(Class, methodNames, supClass, PACKAGE)
        {
            if (!isClass(Class))
                stop("Not a valid class:", Class)
            if (!extends(Class, "externalStorage")) {
                stop("Can not define native storage methods for objects not extending\nexternalStorage")
            }
            ## get the nativeStorageMethodsList object for the superclass
            supMethods <- get(supClass, envir=.nativeMethods)
            if (length(methodNames) > 0) {
                ## Create a list of external pointers - each containing the
                ## address of a native function
                thisMethodsList <- lapply(methodNames,
                                          function(mname)
                                      {
                                          getNativeSymbolInfo(mname,
                                                              PACKAGE)$address
                                      })
                names(thisMethodsList) <-
                    if (length(names(methodNames)) == 0)
                        methodNames
                    else ifelse(names(methodNames) == "",
                                methodNames, names(methodNames))
                .Call("createNativeStorageMethods",
                      new("nativeStorageMethodsList"), thisMethodsList,
                      supMethods, PACKAGE="externalVector")
            } else supMethods
        }

    ## setExternalStorageClass <-
    function(Class, methodNames, ...,
             contains = "externalStorage",
             where = topenv(parent.frame()))
    {
        supClasses <- contains[unlist(lapply(contains,
                                             extends,
                                             "externalStorage"))]
        if (length(supClasses) == 0)
            stop("invalid class definition: must contain externalStorage or its subclass")
        setClass(Class, ..., contains = contains, where = where)
        package <- attr(getClass(Class)@className, "package")
        assign(Class,
               delay(setNativeStorageMethods(Class, methodNames,
                                             supClasses, package),
                     environment()),
               envir=.nativeMethods)
        Class
    }
})

if (!isGeneric("internalType"))
    setGeneric("internalType",
               function(x)
               standardGeneric("internalType"))

setMethod("internalType",
          signature(x="externalStorage"),
          function(x)
      {
          type <- x@type
          if (class(type) %in% c("logical", "integer", "numeric",
                                 "complex", "character")) {
              type
          } else vector("list", 1)
      })

setMethod("initializeResource", "externalStorage",
          function(resource, ptr, size, type, ...)
       {
           resource@ptr <- ptr
           resource@type <- type
           resource@length <- size
           resource
       })

setMethod("getPointer",
          signature(resource="externalStorage"),
          function(resource)
      {
          resource@ptr
      })
setExternalStorageClass("simpleStorage", c("logicalGetElt",
                                           "logicalSetElt",
                                           "integerGetElt",
                                           "integerSetElt",
                                           "numericGetElt",
                                           "numericSetElt",
                                           "complexGetElt",
                                           "complexSetElt",
                                           "imaginaryGetElt",
                                           "imaginarySetElt",
                                           "characterGetElt",
                                           "characterSetElt",
                                           "anyGetElt",
                                           "anySetElt",
                                           "size",
                                           "resize"),
                        contains = "externalStorage")

setMethod("allocatedType",
          signature(resource="simpleStorage"),
          function (resource) 
      {
          internalType(resource)
      })

setMethod("allocatedSize",
          signature(resource="simpleStorage"),
          function(resource)
      {
          resource@length
      })
externalVector <-
    function(type = "logical", length = 0, storageClass =
    "simpleStorage", indirect = FALSE)
{
    if (!is.character(type))
        stop('invalid argument type = ', type)
    type <-
        if (type %in% c("logical", "numeric", "integer", "complex",
                        "character"))
            vector(type, 1)
        else new(type)
    ans <- new("externalVectorWithStorage",
               storage = new(storageClass, type = type,
               length = as.integer(length)))
    if (indirect)
        as(ans, "indirectExternalVector")
    else ans
}

externalNumeric <-
    function(length = 0, storageClass = "simpleStorage", indirect = FALSE)
    externalVector(type = "numeric", length, storageClass, indirect)

externalInteger <-
    function(length = 0, storageClass = "simpleStorage", indirect = FALSE)
    externalVector(type = "integer", length, storageClass, indirect)

externalLogical <-
    function(length = 0, storageClass = "simpleStorage", indirect = FALSE)
    externalVector(type = "logical", length, storageClass, indirect)

externalComplex <-
    function(length = 0, storageClass = "simpleStorage", indirect = FALSE)
    externalVector(type = "complex", length, storageClass, indirect)

externalCharacter <-
    function(length = 0, storageClass = "simpleStorage", indirect = FALSE)
    externalVector(type = "character", length, storageClass, indirect)

externalMatrix <-
    function (data = NA, nrow = 1, ncol = 1, byrow = FALSE,
              dimnames = NULL,
              storageClass = (if (is(data, "externalVectorWithStorage"))
                              class(data@storage) else
                              "simpleStorage"),
              indirect = FALSE)
{
    if (!is(data, "externalVectorWithStorage"))
        data <- as.Rvector(data)
    ansTypeName <- class(defaultElement(data))
    if (missing(nrow)) 
        nrow <- ceiling(length(data)/ncol)
    else if (missing(ncol)) 
        ncol <- ceiling(length(data)/nrow)
    anslen <- nrow*ncol
    if (byrow && anslen > 1)
        stop("option `byrow = TRUE' not implemented")
    datlen <- length(data)
    if (length(data) < 0)
        stop("argument has length zero")
    ## The following is mostly copied from do_matrix in R/src/main/array.c
    if (datlen > 1 && anslen %% datlen != 0) {
	if (((datlen > nrow) && (datlen / nrow) * nrow != datlen) ||
	    ((datlen < nrow) && (nrow / datlen) * datlen != nrow))
	    warning("Replacement length not a multiple of the elements to replace in externalMatrix(...)")
	else if (((datlen > ncol) && (datlen / ncol) * ncol != datlen) ||
		 ((datlen < ncol) && (ncol / datlen) * datlen != ncol))
	    warning("Replacement length not a multiple of the elements to replace in matrix(...)")
    } else if ((datlen > 1) && (anslen == 0)){
        warning("Replacement length not a multiple of the elements to replace in matrix(...)")
    } else if (datlen == 0 && anslen > 0){
        stop("No data to replace in externalMatrix(...)")
    }
    ## now we create the matrix and copy the values to it
    ans <- externalVector(ansTypeName, anslen, storageClass,
                          indirect=indirect)
    ans[] <- data
    dim(ans) <- c(nrow, ncol)
    dimnames(ans) <- dimnames
    ans
}
### externalVector class: Vector objects with external storage.

setVirtualMethod('setNames', signature(object='externalVector'),
                 def=function(object, nm) standardGeneric('setNames'))

setVirtualMethod('setDimNames', signature(object='externalVector'),
                 def=function(object, dnm) standardGeneric('setDimNames'))

setReplaceMethod("names",
                 signature(x="externalVector",
                           value="externalVector"),
                 function(x, value)
             {
                 if (class(defaultElement(value)) != "character")
                     value <- asEach(value, "character")
                 if (length(value) > length(x))
                     stop("Names slot must be the same length as the externalVector")
                 if (length(value) < length(x))
                     length(value) <- length(x)
                 setNames(x, value)
             })

setReplaceMethod("names",
                 signature(x="externalVector"),
                 function(x, value)
             {
                 if (!is.character(value))
                     value <- as.character(value)
                 if (length(value) > length(x))
                     stop("Names slot must be the same length as the externalVector")
                 if (length(value) < length(x))
                     length(value) <- length(x)
                 setNames(x, value)
             })

setReplaceMethod("names",
                 signature(x="externalVector", value = "NULL"),
                 function(x, value)
             {
                 if (is.null(names(x)))
                     x
                 else setNames(x, NULL)
             })

setReplaceMethod("dimnames", signature(x="externalVector",
                                       value="NULL"),
                 function(x, value)
             {
                 if (is.null(dim(x)))
                     stop("dimnames applied to non-array")
                 if (is.null(dimnames(x)))
                     x
                 else setDimNames(x, NULL)
             })

setReplaceMethod("dimnames",
                 signature(x="externalVector",
                           value="list"),
                 function(x, value)
             {
                 if (is.null(d <- dim(x)))
                     stop("dimnames applied to non-array")
                 if (length(value) != length(d))
                     stop("length of dimnames must match that of dims")
                 for (i in seq(along=d)) {
                     if (!is.null(value[[i]])) {
                         this <- value[[i]]
                         if (!is.vector(this) && !is(this, "externalVector"))
                             stop("invalid type for dimname (must be a vector)")
                         if (d[i] != length(this) &&
                             length(this) != 0)
                             stop("length of dimnames[",
                                  i, "] not equal to array extent")
                         value[[i]] <-
                             if (length(this) == 0)
                                 NULL
                             else if (is(this,
                                         "externalVector"))
                                 asEach(this, "character")
                             else as.character(this)
                     }
                 }
                 setDimNames(x, value)
             })

if (!isGeneric("getExternalStorage"))
    setGeneric("getExternalStorage",
               function(x)
               standardGeneric("getExternalStorage"))

setClass("externalVectorWithStorage",
         representation(storage = "externalStorage",
                        Names = "vectorNamesType", # names for the vector
                        Dim = "integerOrNull",     # dimension (may be NULL)
                        DimNames = "listOrNull"),
         contains = "externalVector",
         prototype = list(Names = NULL))

setMethod("initialize",
          signature(.Object="externalVectorWithStorage"),
          function(.Object, ...)
      {
          .Object <- callNextMethod()
          if (class(.Object@storage) == "externalStorage")
              .Object@storage <-
                  new("simpleStorage", type = vector("logical", 1),
                      length = as.integer(0))
          .Object[] <- defaultElement(.Object)
          .Object
      })

setMethod("getExternalStorage",
          signature(x="externalVectorWithStorage"),
          function (x) 
      {
          x@storage
      })

setIs("externalVectorWithStorage", "vectorNamesType")

setMethod("getNativeStorageMethods", "externalVectorWithStorage",
          function(x)
      {
          getNativeStorageMethods(x@storage)
      })

setMethod("show",
          signature(object = "externalVector"),
          function(object)
      {
          objDim <- dim(object)
          cat("External",
              if (is.null(objDim)) "vector"
              else if (length(objDim) == 2) "matrix"
              else "array",
              "object with\n")
          cat("  Class:", class(object), "\n")
          cat("  Storage:", class(getExternalStorage(object)), "\n")
          cat("  Type:", class(defaultElement(object)), "\n")
          if (is.null(objDim) || length(objDim) == 1) {
              n <- length(object)
              cat("  Length:", n, "\n")
              if (n > 0) {
                  if (n > 5) {
                      object <- as.Rvector(force(object[c(1:4, n)]))
                      nm <- names(object)
                      object <-
                          if (is.character(object)) {
                              paste('"', object, '"', sep = '')
                          } else format(object)
                      object <- c(object[1:4], "...", object[5])
                      names(object) <- 
                          if (is.null(nm))
                              c(1:4, '', n)
                          else c(nm[1:4], '', nm[5])
                  } else {
                      object <- as.Rvector(force(object))
                      object <-
                          if (is.character(object))
                              paste('"', object, '"', sep = '')
                          else format(object)
                  }
              }
          } else {
              cat("  Dim:", objDim, "\n")
              if (length(object) > 0) {
                  if (length(objDim) == 2) {
                      i <- if (objDim[1] > 5) c(1:4, objDim[1]) else seq(length=objDim[1])
                      j <- if (objDim[2] > 5) c(1:4, objDim[2]) else seq(length=objDim[2])
                      object <- force(object[i, j, drop = FALSE])[,,drop=FALSE]
                      dnm <- dimnames(object)
                      if (is.null(dnm))
                          dnm <- list(NULL, NULL)
                      if (is.null(dnm[[1]])) {
                          dnm[[1]] <- format(paste('[', i, ',]', sep=''),
                                             justify='right') 
                      }
                      if (is.null(dnm[[2]])) {
                          dnm[[2]] <- paste('[,', j, ']', sep='')
                      }
                      if (length(i) != objDim[1]) {
                          mid <- ceiling(unlist(lapply(object[4, ], nchar))/2)
                          dots <- character(mid)
                          dots[] <- ' '
                          dots[mid] <- '.'
                          dots <- paste(dots, collapse = '')
                          object <- rbind(object[1:4, , drop = FALSE],
                                          matrix(rep(dots, 3*ncol(object)),
                                                 nrow=3),
                                          object[5, ])
                          dnm[[1]] <- c(dnm[[1]][1:4], rep('', 3), dnm[[1]][5])
                      }
                      if (length(j) != objDim[2]) {
                          dnm[[2]] <- c(dnm[[2]][1:4], '', dnm[[2]][5])
                          object <- format(rbind(dnm[[2]],
                                                 cbind(object[, 1:4, drop = FALSE],
                                                       if (length(i) != objDim[1]) {
                                                           rep('...', 8)
                                                       } else {
                                                           rep('...', objDim[1])
                                                       },
                                                       object[, 5])),
                                           justify = 'right')
                          dnm[[2]] <- object[1, ]
                          object <- object[-1, , drop = FALSE]
                      }
                      dimnames(object) <- dnm
                  } else if (length(objDim) > 2) {
                      object <- NULL
                  }
              }
          }
          if (length(object) > 0)
              print(object, quote=FALSE)
      })

setMethod("dim", "externalVectorWithStorage",
          function(x) x@Dim)

setReplaceMethod("dim", signature(x="externalVectorWithStorage",
                                  value="NULL"),
                 function(x, value)
             {
                 if (!is.null(x@Dim)) {
                     if (!is.null(x@DimNames))
                         x@DimNames <- NULL
                     x@Dim <- NULL
                 }
                 names(x) <- NULL
                 x
             })

setReplaceMethod("dim", signature(x="externalVectorWithStorage"),
                 function(x, value)
             {
                 value <- as.integer(value)
                 if (length(value) == 0)
                     stop("dim: Invalid dimension vector")
                 if (length(x) != prod(value))
                     stop("dim<- length of dims do not match the length of object")
                 x@Dim <- value
                 if (!is.null(x@DimNames))
                     x@DimNames <- NULL
                 names(x) <- NULL
                 x
             })

setMethod("dimnames", "externalVectorWithStorage",
          function(x) x@DimNames)

rebind <-
    function(x, length.x = length(x), type = x@storage@type,
             names.x = x@Names, dim.x = x@Dim,
             dimnames.x = x@DimNames, ..., arrayOnly = FALSE)
{
    if (!is(x, "externalVectorWithStorage"))
        stop("x must be from class externalVectorWithStorage")
    if (missing(type))
        typeName <- class(x@storage@type)
    else {
        typeName <- if (is.character(type))
            type
        else class(type)
    }
    if (typeName %in% c("logical", "numeric", "double", "integer", "complex",
                        "character"))
        type <- vector(typeName, 1)
    else if (is.character(type))
        type <- new(type)
    length.x <- as.integer(length.x)
    if (as.logical(arrayOnly)) {
        new(class(x), storage = new(class(x@storage), type=type,
                      length=length.x))
    } else {
        new(class(x), storage = new(class(x@storage), type=type,
                      length=length.x),
            Names = names.x, Dim = dim.x, DimNames = dimnames.x, ...)
    }
}

if (!isGeneric("defaultElement"))
    setGeneric("defaultElement",
               function(x)
               standardGeneric("defaultElement"))

setMethod("defaultElement", "externalVector",
          function(x) getExternalStorage(x)@type)

setMethod("defaultElement", "vector",
          function(x) vector(class(x), 1))

setMethod("length", "externalVectorWithStorage",
          function(x)
          x@storage@length)

setReplaceMethod("length", signature(x = "externalVectorWithStorage"),
                 function(x, value)
             {
                 value <- as.integer(value)
                 if (value != length(x)) {
                     x <- .Call("externalVector_setLength", x,
                                getNativeStorageMethods(x),
                                value,
                                PACKAGE="externalVector")
                 }
                 x
             })

if (!isGeneric("asEach"))
    setGeneric("asEach",
               function(x, type, arrayOnly=FALSE)
               standardGeneric("asEach"))

setMethod("asEach",
          signature(type="character"),
          function (x, type, arrayOnly) 
      {
          if (arrayOnly)
              as.vector(as.Rvector(x), type)
          else {
              saveAttribs <- nonSlotAttributes(x)
              saveAttribs$dim <- dim(x)
              saveAttribs$dimnames <- dim(x)
              saveAttribs$names <- dim(x)
              x <- as.vector(as.Rvector(x), type)
              attributes(x) <- saveAttribs
              x
          }
      })

setMethod("asEach",
          signature(x="externalVectorWithStorage", type = "character"),
          function(x, type, arrayOnly)
      {
          if (defaultElement(x) == type)
              x
          else {
              y <- rebind(x, type = type, arrayOnly=arrayOnly)
              .Call("externalVector_asEach", x,
                    getNativeStorageMethods(x@storage),
                    y, getNativeStorageMethods(y@storage),
                    PACKAGE = "externalVector")
          }
      })

setMethod("names", "externalVectorWithStorage",
          function(x) x@Names)

setReplaceMethod("names",
                 signature(x="externalVectorWithStorage", value = "NULL"),
                 function(x, value)
             {
                 if (!is.null(x@Names))
                     x@Names <- NULL
                 x
             })

setMethod("setNames", signature(object="externalVectorWithStorage"),
          function(object, nm)
      {
          object@Names <- nm
          object
      })

setMethod("setDimNames", signature(object="externalVectorWithStorage"),
          function(object, dnm)
      {
          object@DimNames <- dnm
          object
      })

setMethod("match", signature(x="externalVectorWithStorage"),
          function(x, table, nomatch, incomparables)
      {
          if (missing(nomatch))
              nomatch <- as.integer(NA)
          else nomatch <- as.integer(nomatch)

          n <- length(x)

          ## the class of answer will have the same subclass of
          ## externalVector but with type integer
          ans <- rebind(x, type = integer(1))

          ## split the problem into pieces of length 65536
          nsplits <- n / 65536
          nleft <- n %% 65536
          indx <- 1:65536

          ## match each piece of x
          for (i in seq(length = nsplits)) {
              ans[indx] <- match(x[indx], table, nomatch)
              indx[] <- indx + 65536
          }
          if (nleft > 0) {
              length(indx) <- nleft
              ans[indx] <- match(x[indx], table, nomatch)
          }

          ans
      })

setMethod("match", signature(table="externalVectorWithStorage"),
          function(x, table, nomatch, incomparables)
      {
          if (length(x) == 0)
              return (integer(0))

          n <- length(table)
          ans <- integer(length(x))

          ## This can be done more efficiently (possibly in C)

          ## split the problem in to pieces of length 65536
          nsplits <- n / 65536
          nleft <- n %% 65536
          indx <- 1:65536
          unmatched <- ans

          ## match each piece of x
          for (i in seq(length = nsplits)) {
              ans[unmatched] <- ans + match(x[unmatched], table[indx],
                                            nomatch = 1-indx[1]) + indx[1]-1
              unmatched <- ans == 0
              if (!any(unmatched)) {
                  nleft <- 0
                  break
              }
              indx[] <- indx + 65536
          }
          if (nleft > 0) {
              length(indx) <- nleft
              ans[unmatched] <- ans + match(x[unmatched], table[indx],
                                            nomatch = 1-indx[1]) + indx[1]-1
              unmatched <- ans == 0
          }
          ans[unmatched] <- if (missing(nomatch))
              as.integer(NA)
          else as.integer(nomatch)
          ans
      })

setMethod("[",
          signature(x="externalVectorWithStorage"),
          function(x, i, j, ..., drop)
      {
          if (missing(drop)) {
              drop <- FALSE
              nsubs <- nargs() - 1
          } else {
              nsubs <- nargs() - 2
          }
          if (nsubs == 0)
              nsubs <- 1
          if (nsubs == 1 && missing(i))
              x
          else {
              subs <-
                  if (nsubs == 1) {
                      alist(i=)
                  } else if (nsubs == 2) {
                      alist(i=, j=)
                  } else {
                      alist(i=, j=, ...)
                  }
              if (!missing(i))
                  subs['i'] <- list(i)
              if (!missing(j))
                  subs['j'] <- list(j)
              .Call("externalSubset", x,
                    getNativeStorageMethods(x@storage),
                    if (is(names(x), "externalVectorWithStorage"))
                    getNativeStorageMethods(names(x)@storage)
                    else NULL,
                    SubscriptList(subs, length(x), dim(x),
                                  names(x), dimnames(x), FALSE), drop,
                    PACKAGE="externalVector")
          }
      })

setReplaceMethod("[", signature(x="externalVectorWithStorage",
                                value="externalVectorWithStorage"),
                 function(x, i, j, ..., value)
             {
                 nsubs <- nargs() - 2
                 if (nsubs == 0)
                     nsubs <- 1
                 subs <-
                     if (nsubs == 1) {
                         alist(i=)
                     } else if (nsubs == 2) {
                         alist(i=, j=)
                     } else {
                         alist(i=, j=, ...)
                     }
                 if (!missing(i))
                     subs$i <- i
                 if (!missing(j))
                     subs$j <- j
                 subs <- SubscriptList(subs, length(x), dim(x),
                                       names(x), dimnames(x), TRUE)
                 .Call("externalSubassign", x,
                       getNativeStorageMethods(x@storage),
                       subs, attr(subs, 'newnames'),
                       attr(subs, 'newlength'), value,
                       getNativeStorageMethods(value@storage),
                       PACKAGE="externalVector")
             })

setReplaceMethod("[", "externalVectorWithStorage",
                 function(x, i, j, ..., value)
             {
                 nsubs <- nargs() - 2
                 if (nsubs == 0)
                     nsubs <- 1
                 subs <-
                     if (nsubs == 1) {
                         alist(i=)
                     } else if (nsubs == 2) {
                         alist(i=, j=)
                     } else {
                         alist(i=, j=, ...)
                     }
                 if (!missing(i))
                     subs$i <- i
                 if (!missing(j))
                     subs$j <- j
                 subs <- SubscriptList(subs, length(x), dim(x),
                                       names(x), dimnames(x), TRUE)
                 .Call("externalSubassign", x,
                       getNativeStorageMethods(x@storage),
                       subs, attr(subs, 'newnames'),
                       attr(subs, 'newlength'), value, NULL,
                       PACKAGE="externalVector")
             })

setMethod("is.finite",
          signature(x = "externalVectorWithStorage"),
          function (x)
          is.finite(as.Rvector(x)))

setMethod("is.infinite",
          signature(x = "externalVectorWithStorage"),
          function (x) is.infinite(as.Rvector(x)))

setMethod("is.na",
          signature(x = "externalVectorWithStorage"),
          function (x) is.na(as.Rvector(x)))

setMethod("is.nan",
          signature(x = "externalVectorWithStorage"),
          function (x) is.nan(as.Rvector(x)))
setMethod("as.vector",
          signature(x = "externalVector"),
          function (x, mode) 
      {
          if (missing(mode))
              mode <- "any"
          type <- class(as.vector(defaultElement(x), mode))
          if (type != class(defaultElement(x)))
              x <- asEach(x, type, arrayOnly=TRUE)
          ## strip attributes
          nonSlotAttributes(x) <- NULL
          x
      })

setMethod("as.logical",
          signature(x="externalVector"),
          function (x, ...)
          as.vector(x, "logical"))

setMethod("as.integer",
          signature(x="externalVector"),
          function (x, ...)
          as.vector(x, "integer"))

setMethod("as.numeric",
          signature(x="externalVector"),
          function (x, ...)
          as.vector(x, "numeric"))

setMethod("as.double",
          signature(x="externalVector"),
          function (x, ...)
          as.vector(x, "numeric"))

setMethod("as.single",
          signature(x="externalVector"),
          function (x, ...)
      {
          x <- as.vector(x, "numeric")
          attr(x, "Csingle") <- TRUE
          x
      })

setMethod("as.complex",
          signature(x="externalVector"),
          function (x, ...)
          as.vector(x, "complex"))

setMethod("as.character",
          signature(x="externalVector"),
          function (x) ## Note: does not have ... in generic
          as.vector(x, "character"))

setMethod("as.list",
          signature(x="externalVector"),
          function (x, ...)
          as.vector(x, "list"))

setMethod("as.matrix",
          signature(x = "externalVector"),
          function (x) 
      {
          if (length(dim(x)) != 2) {
              nm <- names(x)
              dim(x) <- c(length(x), 1)
              if (!is.null(nm))
                  dimnames(x) <- list(nm, NULL)
          }
          ## strip attributes
          nonSlotAttributes(x) <- NULL
          x
      })

## coercion to basic R vector data types

if (!isGeneric("as.Rvector"))
    setGeneric("as.Rvector",
               function(x)
               standardGeneric("as.Rvector"))

setMethod("as.Rvector",
          "logical",
          function (x) x)

setMethod("as.Rvector",
          "integer",
          function (x) x)

setMethod("as.Rvector",
          "numeric",
          function (x) x)

setMethod("as.Rvector",
          "complex",
          function (x) x)

setMethod("as.Rvector",
          "character",
          function (x) x)

setMethod("as.Rvector",
          "list",
          function (x) x)

setMethod("as.Rvector", "externalVector",
          function(x) as.Rvector(force(x)))

setMethod("as.Rvector",
          signature(x="externalVectorWithStorage"),
          function (x) 
      {
          Dim <- dim(x)
          nDim <- length(Dim)
          Names <- names(x)
          saveAttribs <- nonSlotAttributes(x)
          if (nDim > 0) {
              DimNames <- dimnames(x)
              args <- vector('list', nDim+1)
              args[[1]] <- x
              args[-1] <- new("allIndex")
              x <- do.call('[', args)
              dim(x) <- Dim
              dimnames(x) <- DimNames
          } else {
              x <- x[new("allIndex")]
          }
          names(x) <- Names
          nonSlotAttributes(x) <- saveAttribs
          x
      })

setMethod("asEach",
          signature(x="externalVector", type = "character"),
          function(x, type, arrayOnly)
      {
          type <- class(as.vector(defaultElement(x), type))
          if (class(defaultElement(x)) == type)
              x
          else {
              x <- force(x)
              callNextmethod()
          }
      })

setMethod("as.data.frame",
          signature(x="externalVector"),
          function(x, row.names, optional)
          as.data.frame(as.Rmatrix(x), row.names, optional))

setMethod("c", 
          signature(x="externalVector"),
          function(x, ..., recursive)
      {
          c(as.Rvector(x),
            c(..., recursive=recursive), recursive=recursive)
      })

setAs("externalVector", "vector",
      function(from, to) as.Rvector(from))

setAs("externalVector", "matrix",
      function(from, to) as.Rvector(as.matrix(from)))

setAs("vector", "externalVector",
      function(from, to)
  {
      to <- externalVector(class(from), length=length(from))
      to[] <- from
      d <- dim(to) <- dim(from)
      if (!is.null(d))
          dimnames(to) <- dimnames(from)
      nm <- names(from)
      if (!is.null(nm))
          names(to) <- nm
      to
  })

setAs("matrix", "externalVector",
      function(from, to)
  {
      to <- externalMatrix(from,
                           nrow = nrow(from),
                           ncol = ncol(from),
                           dimnames = dimnames(from))
      nm <- names(from)
      if (!is.null(nm))
          names(to) <- nm
      to
  })
setClass("matrixIndex",
         representation(row="vectorIndex",
                        col="vectorIndex",
                        origdim="integer",
                        Names="vectorNamesType"),
         contains="externalVector",
         prototype=list(origdim=as.integer(c(0, 0))))

setMethod("any.na",
          signature(x="matrixIndex"),
          function (x) 
      {
          any.na(x@row) || any.na(x@col)
      })

setMethod("all.na",
          signature(x="matrixIndex"),
          function (x) 
      {
          all.na(x@row) && all.na(x@col)
      })

setMethod("length",
          signature(x = "matrixIndex"),
          function (x) 
      {
          as.integer(length(x@row)*length(x@col))
      })

setMethod("names",
          signature(x="matrixIndex"),
          function (x) x@Names)

setMethod("setNames",
          signature(object = "matrixIndex"),
          function (object, nm) 
      {
          object@Names <- nm
          object
      })

setMethod("dim",
          signature(x = "matrixIndex"),
          function (x) 
      {
          as.integer(c(length(x@row), length(x@col)))
      })

setReplaceMethod("dim",
                 signature(x = "noneIndex"),
                 function (x, value) 
             {
                 value <- as.integer(value)
                 if (length(value) == 0)
                     x
                 else if (length(value) != 2 || prod(value) != 0)
                     stop("invalid dimension")
                 else {
                     m <- value[1]
                     n <- value[2]
                     origdim <- c(m, n)
                     new("matrixIndex",
                         row = if (m == 0)
                         new("noneIndex")
                         else makeSlice(0, m, 1),
                         col = if (n == 0) new("noneIndex")
                         else makeSlice(0, n, 1),
                         origdim = origdim)
                 }
             })

setReplaceMethod("dim",
                 signature(x = "allIndex"),
                 function (x, value) 
             {
                 value <- as.integer(value)
                 if (length(value) == 0)
                     x
                 else if (length(value) != 2 || prod(value) < 0)
                     stop("invalid dimension")
                 else {
                     m <- value[1]
                     n <- value[2]
                     new("matrixIndex",
                         row = if (m == 0)
                         new("noneIndex")
                         else makeSlice(0, m, 1),
                         col = if (n == 0)
                         new("noneIndex")
                         else makeSlice(0, n, 1),
                         origdim = value)
                 }
             })

setReplaceMethod("dim",
                 signature(x = "NAIndex"),
                 function (x, value) 
             {
                 value <- as.integer(value)
                 if (length(value) == 0)
                     x
                 else if (length(value) != 2 || prod(value) != length(x))
                     stop("invalid dimension")
                 else {
                     m <- value[1]
                     n <- value[2]
                     new("matrixIndex",
                         row = if (m == 0)
                         new("noneIndex")
                         else new("NAIndex", Length=m),
                         col = if (n == 0)
                         new("noneIndex")
                         else new("NAIndex", Length=n),
                         origdim = value)
                 }
             })

setReplaceMethod("dim",
                 signature(x = "sliceIndex"),
                 function (x, value) 
             {
                 value <- as.integer(value)
                 if (length(value) == 0)
                     x
                 else if (length(value) != 2 || prod(value) !=
                          length(x) || x@content[3] != 1)
                     stop("invalid dimension")
                 else {
                     m <- value[1]
                     n <- value[2]
                     start <- x@content[1]
                     new("matrixIndex",
                         row = if (m == 0)
                         new("noneIndex")
                         else makeSlice(start %% m,
                                        m, 1),
                         col = if (n == 0)
                         new("noneIndex")
                         else makeSlice(start / m,
                                        n, 1),
                         origdim = value)
                 }
             })

setReplaceMethod("dim",
                 signature(x = "positiveIndex"),
                 function (x, value) 
             {
                 x <- as(x, "sliceIndex")
                 dim(x) <- value
                 x
             })

setMethod("dimnames",
          signature(x = "matrixIndex"),
          function (x) 
      {
          nm1 <- names(x@row)
          nm2 <- names(x@col)
          if (is.null(nm1) && is.null(nm2))
              NULL
          else list(nm1, nm2)
      })

setMethod("setDimNames",
          signature(object = "matrixIndex"),
          function (object, dnm) 
      {
          if (is.null(dnm)) {
              names(object@row) <- NULL
              names(object@col) <- NULL
          } else {
              if (!is.null(dnm[[1]]))
                  names(object@row) <- dnm[[1]]
              if (!is.null(dnm[[2]]))
                  names(object@col) <- dnm[[2]]
          }
          object
      })

setMethod("[[",
          signature(x = "matrixIndex", i = "numeric", j = "missing"),
          function (x, i, j, ...) 
      {
          i <- as.integer(i)
          if (i == 1)
              x@row
          else if (i == 2)
              x@col
          else NULL
      })

setMethod("[",
          signature(x = "matrixIndex"),
          function (x, i, j, ..., drop) 
      {
          if (missing(drop)) {
              drop <- FALSE
              nsubs <- nargs() - 1
          } else {
              nsubs <- nargs() - 2
          }
          if (nsubs == 0)
              nsubs <- 1
          if (nsubs == 1 && missing(i))
              return(x)
          subs <-
              if (nsubs == 1) {
                  alist(i=)
              } else if (nsubs == 2) {
                  alist(i=, j=)
              } else {
                  alist(i=, j=, ...)
              }
          if (!missing(i))
              subs$i <- i
          if (!missing(j))
              subs$j <- j
          subs <- SubscriptList(subs, length(x), dim(x),
                                x@Names, dimnames(x), FALSE)
          n <- x@origdim[1]
          if (length(subs) == 1) {
              nm <- x@Names
              x <- .Call('matrixToVectorIndex', x@row, length(x@row),
                         x@col, length(x@col), x@origdim,
                         subs[[1]], length(subs[[1]]),
                         PACKAGE='externalVector')
              if (length(nm) > 0)
                  names(x) <- Subset(nm, x)
              x
          } else {
              x@row <- x@row[subs[[1]]]
              x@col <- x@col[subs[[2]]]
              if (drop &&
                  (length(x@row) == 1 || length(x@col) == 1)) {
                  if (length(x@row) == 1) {
                      if (length(x@col) == 1)
                          as(x@row[1]+n*x@col[1], 'positiveIndex')
                      else x@row[1]+n*x@col
                  } else x@row+n*x@col[1]
              } else {
                  x@Names <- NULL
                  x
              }
          }
      })

setAs("matrixIndex", "vectorIndex",
      function(from, to)
  {
      ans <- .Call('matrixToVectorIndex', from@row, length(from@row),
                   from@col, length(from@col), from@origdim,
                   new("allIndex"), length(from),
                   PACKAGE='externalVector')
      if (!is.null(names(from)))
          names(ans) <- from
      ans
  })

setMethod("Subset",
          signature(i="matrixIndex"),
          function (x, i) 
      {
          dim(x) <- i@origdim
          Subset2(x, i@row, i@col)
      })
setClass("indirectExternalVector",
         representation(actual="externalVector",
                        index = "vectorIndex"),
         contains = "externalVector")

if (!isClass("exprMatrix"))
    setClassUnion("exprMatrix", c("matrix"))

## Class union for the exprs and se.exprs slots of exprSet
if (!isClass("exprMatrix"))
    setClassUnion("exprMatrix", c("matrix"))

setClass("indirectExternalMatrix",
         representation(actual="externalVector",
                        index = "matrixIndex"),
         contains = c("externalVector"))

setIs("indirectExternalMatrix", "exprMatrix")

setMethod("getExternalStorage",
          signature(x="indirectExternalVector"),
          function (x) getExternalStorage(x@actual))

if (!isGeneric("force")) {
    setGeneric("force")
}

setMethod("names", "indirectExternalVector",
          function(x) names(x@index))

setMethod("length",
          signature(x = "indirectExternalVector"),
          function (x) 
      {
          length(x@index)
      })

setMethod("force",
          signature(x="indirectExternalVector"),
          function (x)
      {
          ans <- force(x@actual)
          if (is(x@index, "sliceIndex") && x@index@content == c(0,
                                                   length(ans), 1.0)) {
              dim(ans) <- NULL
          } else {
              ans <- Subset(ans, x@index)
          }
          names(ans) <- names(x)
          ans
      })

setReplaceMethod("names",
                 signature(x="indirectExternalVector", value = "NULL"),
                 function(x, value)
             {
                 names(x@index) <- NULL
                 x
             })

setMethod("setNames",
          signature(object="indirectExternalVector"),
          function(object, nm)
      {
          names(object@index) <- nm
          object
      })

setReplaceMethod("length",
                 signature(x = "indirectExternalVector"),
                 function (x, value) 
             {
                 stop("cannot modify length of indirect external vector")
             })

setMethod("dim",
          signature(x = "indirectExternalVector"),
          function (x) NULL)

setReplaceMethod("dim", signature(x="indirectExternalVector"),
                 function(x, value)
             {
                 value <- as.integer(value)
                 if (length(value) == 0)
                     stop("dim: Invalid dimension vector")
                 if (length(x@actual) != prod(value))
                     stop("dim<- length of dims do not match the length of original object")
                 indx <- x@index
                 dim(indx) <- value
                 indx@origdim <- value
                 new("indirectExternalMatrix",
                     actual=x@actual,
                     index=indx)
             })

setMethod("dimnames",
          signature(x = "indirectExternalVector"),
          function (x) NULL)

setReplaceMethod("dimnames",
                 signature(x = "indirectExternalVector"),
                 function (x, value) 
                 stop("dimnames applied to non-array"))

setMethod("[",
          signature(x = "indirectExternalVector", i = "missing",
                    j = "missing"),
          function (x, i, j, ..., drop) x)

setMethod("[",
          signature(x = "indirectExternalVector", j = "missing"),
          function (x, i, j, ..., drop) 
      {
          i <- x@index[i]
          if (any.na(i))
              stop('subscript out of bounds')
          new("indirectExternalVector",
              actual=x@actual,
              index=i)
      })

setReplaceMethod("[",
                 signature(x = "indirectExternalVector", i = "missing",
                           j = "missing"),
                 function (x, i, j, ..., value) 
             {
                 x@actual[x@index] <- value
                 x
             })

setReplaceMethod("[",
                 signature(x = "indirectExternalVector", j = "missing"),
                 function (x, i, j, ..., value) 
             {
                 i <- x@index[i]
                 if (any.na(i))
                     stop('subscript out of bounds')
                 ans <- new("indirectExternalVector",
                            actual=x@actual,
                            index=i)
                 ans@actual[i] <- value
                 ans
             })

setMethod("getExternalStorage",
          signature(x="indirectExternalMatrix"),
          function (x) getExternalStorage(x@actual))

setMethod("force",
          signature(x="indirectExternalMatrix"),
          function (x)
      {
          ans <- force(x@actual)
          dim(ans) <- x@index@origdim
          if (!(is(x@index@row, "sliceIndex") && x@index@row@content == c(0,
                                                             nrow(ans),
                                                             1.0) &&
                is(x@index@col, "sliceIndex") && x@index@col@content == c(0,
                                                             ncol(ans), 1.0))) {
              ans <- Subset(ans, x@index)
          }
          dim(ans) <- dim(x)
          dimnames(ans) <- dimnames(x)
          names(ans) <- names(x)
          ans
      })

setMethod("length",
          signature(x = "indirectExternalMatrix"),
          function (x) 
      {
          length(x@index)
      })

setReplaceMethod("length",
                 signature(x = "indirectExternalMatrix"),
                 function (x, value) 
             {
                 stop("cannot modify length of indirect external matrix")
             })

setMethod("dim", "indirectExternalMatrix",
          function(x) dim(x@index))

setReplaceMethod("dim", signature(x="indirectExternalMatrix",
                                  value="NULL"),
                 function(x, value)
             {
                 new("indirectExternalVector",
                     actual=x@actual,
                     index=as(x@index, "vectorIndex"))
             })

setReplaceMethod("dim", signature(x="indirectExternalMatrix"),
                 function(x, value)
             {
                 dim(x@index) <- value
                 x
             })

setMethod("names",
          signature(x="indirectExternalMatrix"),
          function (x) 
      {
          names(x@index)
      })

setReplaceMethod("names",
          signature(x="indirectExternalMatrix"),
          function(x, value)
      {
          names(x@index) <- value
          x
      })

setMethod("dimnames",
          signature(x="indirectExternalMatrix"),
          function (x) 
      {
          dimnames(x@index)
      })

setReplaceMethod("dimnames",
                 signature(x="indirectExternalMatrix"),
                 function(x, value)
             {
                 dimnames(x@index) <- value
                 x
             })

setMethod("[",
          signature(x = "indirectExternalMatrix", i = "missing",
                    j = "missing", drop = "missing"),
          function (x, i, j, ..., drop)
      {
          if (nargs() < 3)
              new("indirectExternalVector",
                  actual=x@actual,
                  index=as(x@index, "vectorIndex"))
          else x
      })

setMethod("[",
          signature(x = "indirectExternalMatrix", i = "missing",
                    j = "missing", drop = "logical"),
          function (x, i, j, ..., drop)
      {
          if (nargs() < 4)
              new("indirectExternalVector",
                  actual=x@actual,
                  index=as(x@index, "vectorIndex"))
          else if (drop) {
              i <- x@index[,,drop=TRUE]
              if (is(i, 'matrixIndex')) {
                  x@index <- i
                  x
              } else {
                  new("indirectExternalVector",
                      actual=x@actual,
                      index=i)
              }
          } else x
      })

setMethod("[",
          signature(x = "indirectExternalMatrix",
                    i = "missing", drop = "missing"),
          function (x, i, j, ..., drop) 
      {
          if (nargs() != 3)
              stop('incorrect number of subscripts')
          callGeneric(x, new("allIndex"), j, drop=FALSE)
      })

setMethod("[",
          signature(x = "indirectExternalMatrix",
                    i = "missing", drop = "logical"),
          function (x, i, j, ..., drop) 
      {
          if (nargs() != 4)
              stop('incorrect number of subscripts')
          callGeneric(x, new("allIndex"), j, drop=drop)
      })

setMethod("[",
          signature(x = "indirectExternalMatrix",
                    j = "missing", drop = "missing"),
          function (x, i, j, ..., drop) 
      {
          if (nargs() == 2)
              callGeneric(x, i, drop=FALSE)
          else if (nargs() == 3)
              callGeneric(x, i, new("allIndex"), drop=FALSE)
          else stop('incorrect number of subscripts')
      })

setMethod("[",
          signature(x = "indirectExternalMatrix",
                    j = "missing", drop = "logical"),
          function (x, i, j, ..., drop) 
      {
          if (nargs() == 4)
              callGeneric(x, i, new("allIndex"), drop=drop)
          else if (nargs() == 3) {
              i <- x@index[i]
              if (length(i) == 0)
                  force(x@actual)[i]
              else if (any.na(i))
                  stop('subscript out of bounds')
              else {
                  ans <- new("indirectExternalVector",
                             actual=x@actual,
                             index=i)
                  if (!is.null(names(x)))
                      names(ans) <- Subset(names(x), i)
                  ans
              }
          } else stop('incorrect number of subscripts')
      })

setMethod("[",
          signature(x = "indirectExternalMatrix",
                    drop = "missing"),
          function (x, i, j, ..., drop) 
      {
          if (nargs() != 3)
              stop('incorrect number of subscripts')
          callGeneric(x, i, j, drop=FALSE)
      })

setMethod("[",
          signature(x = "indirectExternalMatrix"),
          function (x, i, j, ..., drop) 
      {
          if (nargs() != 4)
              stop('incorrect number of subscripts')
          i <- x@index[i, j, drop = drop]
          if (length(i) == 0) {
              if (is(i, "matrixIndex"))
                  force(x@actual)[i[[1]], i[[2]], drop=drop]
              else force(x@actual)[i, drop=drop]
          } else if (any.na(i))
              stop('subscript out of bounds')
          else if (is(i, 'matrixIndex')) {
              x@index <- i
              x
          } else {
              new("indirectExternalVector",
                  actual=x@actual,
                  index=i)
          }
      })

setReplaceMethod("[",
                 signature(x = "indirectExternalMatrix", i = "missing",
                           j = "missing"),
                 function (x, i, j, ..., value) 
             {
                 if (nargs() > 4)
                     stop('incorrect number of subscripts')
                 if (nargs() == 4)
                     x@actual[x@index[[1]], x@index[[2]]] <- value
                 else x@actual[x@index] <- value
                 x
             })

setReplaceMethod("[",
                 signature(x = "indirectExternalMatrix"),
                 function (x, i, j, ..., value) 
             {
                 nsubs <- nargs() - 2
                 if (nsubs == 0)
                     nsubs <- 1
                 if (nsubs == 1) {
                     i <- if (missing(i)) x@index else x@index[i]
                     if (any.na(i))
                         stop('subscript out of bounds')
                     x@actual[i] <- value
                 } else if (nsubs == 2) {
                     i <- if (missing(i)) x@index[[1]] else x@index[[1]][i]
                     if (any.na(i))
                         stop('subscript out of bounds')
                     j <- if (missing(j)) x@index[[2]] else x@index[[2]][j]
                     if (any.na(j))
                         stop('subscript out of bounds')
                     x@actual[i, j] <- value
                 } else stop('incorrect number of subscripts')
                 x
             })

setAs("externalVector", "indirectExternalVector",
          function (from, to) 
      {
          to <- new("indirectExternalVector", actual=from,
                    index = makeSlice(0, length(from), 1))
          if (!is.null(names(from)))
              names(to) <- names(from)
          to
      })

setAs("externalVector", "indirectExternalMatrix",
          function (from, to) 
      {
          if (length(dim(from)) == 2) {
              to <- new("indirectExternalMatrix", actual=from,
                        index = new("matrixIndex",
                                    row=makeSlice(0, nrow(from), 1),
                                    col=makeSlice(0, ncol(from), 1),
                                    origdim=dim(from)))
              if (!is.null(dimnames(from)))
                  dimnames(to) <- dimnames(from)
              if (!is.null(names(from)))
                  names(to) <- names(from)
          } else {
              to <- new("indirectExternalVector", actual=from,
                        index = makeSlice(0, length(from), 1))
              if (!is.null(names(from)))
                  names(to) <- names(from)
          }
          to
      })

setAs( "indirectExternalVector", "vector",
      function(from, to) as.vector(from))

setAs( "indirectExternalVector", "matrix",
      function(from, to) as.matrix(from))

setAs( "indirectExternalMatrix", "vector",
      function(from, to) as.vector(from))

setAs( "indirectExternalMatrix", "matrix",
      function(from, to) as.matrix(from))

setAs("vector", "indirectExternalVector",
      function(from, to)
  {
      as(as(from, "externalVector"), "indirectExternalVector")
  })

setAs("vector", "indirectExternalMatrix",
      function(from, to)
  {
      as(externalMatrix(from), "indirectExternalMatrix")
  })

setAs("matrix", "indirectExternalMatrix",
      function(from, to)
  {
      as(as(from, "externalVector"), "indirectExternalMatrix")
  })

setMethod("Ops",
          signature(e2="indirectExternalVector"),
          function(e1, e2)
      {
          callGeneric(e1, force(e2))
      })

setMethod("Ops",
          signature(e1="indirectExternalVector"),
          function(e1, e2)
      {
          callGeneric(force(e1), e2)
      })

setMethod("Ops",
          signature(e1="indirectExternalVector",
                    e2="indirectExternalVector"),
          function(e1, e2)
      {
          callGeneric(force(e1), force(e2))
      })

setMethod("Complex",
          signature(z="indirectExternalVector"),
          function(z)
      {
          callGeneric(force(z))
      })

setMethod("Math",
          signature(x="indirectExternalVector"),
          function(x)
      {
          callGeneric(force(x))
      })

setMethod("Ops",
          signature(e2="indirectExternalMatrix"),
          function(e1, e2)
      {
          callGeneric(e1, force(e2))
      })

setMethod("Ops",
          signature(e1="indirectExternalMatrix"),
          function(e1, e2)
      {
          callGeneric(force(e1), e2)
      })

setMethod("Ops",
          signature(e1="indirectExternalMatrix",
                    e2="indirectExternalMatrix"),
          function(e1, e2)
      {
          callGeneric(force(e1), force(e2))
      })

setMethod("Ops",
          signature(e1="indirectExternalMatrix",
                    e2="indirectExternalVector"),
          function(e1, e2)
      {
          callGeneric(force(e1), force(e2))
      })

setMethod("Ops",
          signature(e1="indirectExternalVector",
                    e2="indirectExternalMatrix"),
          function(e1, e2)
      {
          callGeneric(force(e1), force(e2))
      })

setMethod("Complex",
          signature(z="indirectExternalMatrix"),
          function(z)
      {
          callGeneric(force(z))
      })

setMethod("Math",
          signature(x="indirectExternalMatrix"),
          function(x)
      {
          callGeneric(force(x))
      })

setMethod("log",
          signature(x="indirectExternalVector"),
          function(x, base)
      {
          callGeneric(force(x), base)
      })

setMethod("log",
          signature(base="indirectExternalVector"),
          function(x, base)
      {
          callGeneric(x, force(base))
      })

setMethod("log",
          signature(x="indirectExternalVector",
                    base="indirectExternalVector"),
          function(x, base)
      {
          callGeneric(force(x), force(base))
      })

setMethod("log",
          signature(x="indirectExternalMatrix"),
          function(x, base)
      {
          callGeneric(force(x), base)
      })

setMethod("log",
          signature(base="indirectExternalMatrix"),
          function(x, base)
      {
          callGeneric(x, force(base))
      })

setMethod("log",
          signature(x="indirectExternalMatrix",
                    base="indirectExternalMatrix"),
          function(x, base)
      {
          callGeneric(force(x), force(base))
      })

setMethod("log",
          signature(x="indirectExternalVector",
                    base="indirectExternalMatrix"),
          function(x, base)
      {
          callGeneric(force(x), force(base))
      })

setMethod("log",
          signature(x="indirectExternalMatrix",
                    base="indirectExternalVector"),
          function(x, base)
      {
          callGeneric(force(x), force(base))
      })

setMethod("log",
          signature(x="indirectExternalVector",
                    base="missing"),
          function(x, base)
      {
          callGeneric(force(x))
      })

setMethod("log",
          signature(x="indirectExternalMatrix",
                    base="missing"),
          function(x, base)
      {
          callGeneric(force(x))
      })

setMethod("is.finite",
          signature(x = "indirectExternalMatrix"),
          function (x) is.finite(force(x)))

setMethod("is.infinite",
          signature(x = "indirectExternalMatrix"),
          function (x) is.infinite(force(x)))

setMethod("is.na",
          signature(x = "indirectExternalMatrix"),
          function (x) is.na(force(x)))

setMethod("is.nan",
          signature(x = "indirectExternalMatrix"),
          function (x) is.nan(force(x)))

setMethod("any.na",
          signature(x = "indirectExternalMatrix"),
          function (x) any.na(x@index) || any.na(force(x)))

setMethod("all.na",
          signature(x = "indirectExternalMatrix"),
          function (x) all.na(force(x)))

setMethod("is.finite",
          signature(x = "indirectExternalVector"),
          function (x) is.finite(force(x)))

setMethod("is.infinite",
          signature(x = "indirectExternalVector"),
          function (x) is.infinite(force(x)))

setMethod("is.na",
          signature(x = "indirectExternalVector"),
          function (x) is.na(force(x)))

setMethod("is.nan",
          signature(x = "indirectExternalVector"),
          function (x) is.nan(force(x)))

setMethod("any.na",
          signature(x = "indirectExternalVector"),
          function (x) any.na(x@index) || any.na(force(x)))

setMethod("all.na",
          signature(x = "indirectExternalVector"),
          function (x) all.na(force(x)))
setMethod("Ops",
          signature(e1="externalVectorWithStorage"),
          function(e1, e2)
      {
          callGeneric(e1, as(e2, "externalVector"))
      })

setMethod("Ops",
          signature(e2="externalVectorWithStorage"),
          function(e1, e2)
      {
          callGeneric(as(e1, "externalVector"), e2)
      })

setMethod("Arith",
          signature(e1="externalVectorWithStorage",
                    e2="externalVectorWithStorage"),
          function(e1, e2)
      {
          ans <- .Call("externalVectorBinaryArith", .Generic,
                       e1, getNativeStorageMethods(e1@storage),
                       e2, getNativeStorageMethods(e2@storage),
                       PACKAGE="externalVector")
          if (is(ans, "externalStorage")) {
              n <- ans@length
              if (length(e1) == n) {
                  e1@storage <- ans
                  e1
              } else {
                  e2@storage <- ans
                  e2
              }
          } else {
              ans
          }
      })

setMethod("Arith",
          signature(e1="externalVectorWithStorage",
                    e2="missing"),
          function(e1, e2)
      {
          if (.generic == '+')
              e1
          else {
              type <- e1@storage@type
              if (class(type) == "logical")
                  type <- integer(1)
              ans <- rebind(e1, type = type)
              .Call("externalVectorBinaryArith",
                    e1, getNativeStorageMethods(e1@storage),
                    ans, getNativeStorageMethods(ans@storage),
                    PACKAGE="externalVector")
          }
      })

setMethod("Compare",
          signature(e1="externalVectorWithStorage",
                    e2="externalVectorWithStorage"),
          function(e1, e2)
      {
          ans <- .Call("externalVectorCompare",
                       .Generic,
                       e1, getNativeStorageMethods(e1@storage),
                       e2, getNativeStorageMethods(e2@storage),
                       PACKAGE="externalVector")
          if (is(ans, "externalStorage")) {
              n <- ans@length
              if (length(e1) == n) {
                  e1@storage <- ans
                  e1
              } else {
                  e2@storage <- ans
                  e2
              }
          } else {
              ans
          }
      })

setMethod("Complex",
          signature(z="externalVectorWithStorage"),
          function(z)
      {
          ans <- rebind(z,
                        type = if (.Generic == 'Conj' &&
                                   class(x@storage@type) == "complex")
                               complex(1) else numeric(1))
          .Call("externalVectorComplex", .Generic,
                z, getNativeStorageMethods(z@storage),
                ans, getNativeStorageMethods(ans@storage),
                PACKAGE="externalVector")
      })

### Ugly hack to create the generics for max, min etc. which have
### ... as their first argument in base.
for (.nm in c("max", "min", "range", "any", "all",
             "prod", "sum")) {
    if (!isGeneric(.nm)) {
        f <- function(x, ..., na.rm = FALSE)
            .Internal(sum(x, ..., na.rm = na.rm))
        body(f)[[2]][[1]] <- as.name(.nm)
        assign(.nm, f)
        rm(f)
        setGeneric(.nm)
    }
    local({
        .Generic <- .nm
        setMethod(.nm, signature(x="externalVectorWithStorage"),
                  function(x, ..., na.rm)
              {
                  if (length(list(...)) > 0) {
                      callGeneric(force(x),
                                  callGeneric(..., na.rm=na.rm),
                                  na.rm=na.rm)
                  } else callGeneric(force(x), na.rm=na.rm)
              })

        setMethod(.nm, signature(x="externalVectorWithStorage"),
                  function(x, ..., na.rm)
              {
                  na.rm <- as.logical(na.rm)[1]
                  ## We use .External as we have no idea how many
                  ## elements are there in the ... argument.
                  ans <- .External('externalVectorSummary',
                                   .Generic, na.rm, x, ...,
                                   PACKAGE='externalVector')
                  ## The answer we get above is a list with the
                  ## externalVectorWithStorage arguments evaluated
                  ## and at the end. If such a argument was of
                  ## length zero, it is returned unchanged.
                  ## Other arguments are pushed to the front
                  ## (with those with a class pushed in the front so
                  ## that if there is a Summary method for the
                  ## firt one, it will be used next.)
                  
                  ## Now we create a call that will apply the
                  ## summary method to the elements of the list ANS.
                  mcall <- call('callGeneric')
                  if (length(ans) > 0)
                      mcall[seq(from=2, length=length(ans))] <- ans
                  mcall$na.rm <- na.rm
                  eval(mcall)
              })
        setMethod(.nm,
                  signature(x="indirectExternalMatrix"),
                  function(x, ..., na.rm)
              {
                  callGeneric(force(x), !missing(na.rm) && na.rm)
              })
        setMethod(.nm,
                  signature(x="indirectExternalVector"),
                  function(x, ..., na.rm)
              {
                  callGeneric(force(x), !missing(na.rm) && na.rm)
              })
    })
}
rm(.nm)

setMethod("Math",
          signature(x="externalVectorWithStorage"),
          function(x)
      {
          if (.Generic == 'abs') {
              ans <- rebind(x,
                            type = numeric(1))
              return(.Call("externalVectorComplex",
                           .Generic,
                           x, getNativeStorageMethods(x@storage),
                           ans, getNativeStorageMethods(ans@storage),
                           PACKAGE="externalVector"))
          }
          ans <-
              rebind(x,
                     type = if (class(x@storage@type) == "complex") x@storage@type
                     else numeric(1))
          .Call("externalVectorMath", .Generic,
                x, getNativeStorageMethods(x@storage),
                ans, getNativeStorageMethods(ans@storage),
                PACKAGE="externalVector")
      })
if (!isGeneric("colSums"))
    setGeneric("colSums",
               function(x, na.rm=FALSE, dims=1)
               standardGeneric("colSums"))

setMethod("colSums",
          signature(x="externalVectorWithStorage",
                    na.rm="missing",
                    dims = "missing"),
          function(x, na.rm, dims)
      {
          callGeneric(x, na.rm=FALSE, dims=1)
      })

setMethod("colSums",
          signature(x="externalVectorWithStorage",
                    na.rm="missing"),
          function(x, na.rm, dims)
      {
          callGeneric(x, na.rm=FALSE, dims=dims)
      })

setMethod("colSums",
          signature(x="externalVectorWithStorage",
                    dims="missing"),
          function(x, na.rm, dims)
      {
          callGeneric(x, na.rm=na.rm, dims=1)
      })

setMethod("colSums",
          signature(x="externalVectorWithStorage",
                    na.rm="logical",
                    dims="numeric"),
          function(x, na.rm, dims)
      {
          if (length(dn <- dim(x)) < 2)
              stop("`x' must be an external array of at least two dimensions")
          if (dims < 1 || dims > length(dn) - 1) 
              stop("invalid `dims'")
          n <- prod(dn[1:dims])
          dn <- dn[-(1:dims)]
          x <- .Call("colSums", x, getNativeStorageMethods(x),
                     n, prod(dn), na.rm,
                     PACKAGE="externalVector")
          if (length(dn) > 1) {
              dim(x) <- dn
              dimnames(x) <- dimnames(x)[-(1:dims)]
          } else names(x) <- dimnames(x)[[dims + 1]]
          x
      })

if (!isGeneric("colMeans"))
    setGeneric("colMeans",
               function(x, na.rm, dims)
               standardGeneric("colMeans"))

setMethod("colMeans",
          signature(x="externalVectorWithStorage",
                    na.rm="missing",
                    dims = "missing"),
          function(x, na.rm, dims)
      {
          callGeneric(x, na.rm=FALSE, dims=1)
      })

setMethod("colMeans",
          signature(x="externalVectorWithStorage",
                    na.rm="missing"),
          function(x, na.rm, dims)
      {
          callGeneric(x, na.rm=FALSE, dims=dims)
      })

setMethod("colMeans",
          signature(x="externalVectorWithStorage",
                    dims="missing"),
          function(x, na.rm, dims)
      {
          callGeneric(x, na.rm=na.rm, dims=1)
      })

setMethod("colMeans",
          signature(x="externalVectorWithStorage",
                    na.rm="logical",
                    dims="numeric"),
          function(x, na.rm, dims)
      {
          if (length(dn <- dim(x)) < 2)
              stop("`x' must be an external array of at least two dimensions")
          if (dims < 1 || dims > length(dn) - 1) 
              stop("invalid `dims'")
          n <- prod(dn[1:dims])
          dn <- dn[-(1:dims)]
          x <- .Call("colMeans", x, getNativeStorageMethods(x),
                     n, prod(dn), na.rm,
                     PACKAGE="externalVector")
          if (length(dn) > 1) {
              dim(x) <- dn
              dimnames(x) <- dimnames(x)[-(1:dims)]
          } else names(x) <- dimnames(x)[[dims + 1]]
          x
      })

if (!isGeneric("rowSums"))
    setGeneric("rowSums",
               function(x, na.rm, dims)
               standardGeneric("rowSums"))

setMethod("rowSums",
          signature(x="externalVectorWithStorage",
                    na.rm="missing",
                    dims = "missing"),
          function(x, na.rm, dims)
      {
          callGeneric(x, na.rm=FALSE, dims=1)
      })

setMethod("rowSums",
          signature(x="externalVectorWithStorage",
                    na.rm="missing"),
          function(x, na.rm, dims)
      {
          callGeneric(x, na.rm=FALSE, dims=dims)
      })

setMethod("rowSums",
          signature(x="externalVectorWithStorage",
                    dims="missing"),
          function(x, na.rm, dims)
      {
          callGeneric(x, na.rm=na.rm, dims=1)
      })

setMethod("rowSums",
          signature(x="externalVectorWithStorage",
                    na.rm="logical",
                    dims="numeric"),
          function(x, na.rm, dims)
      {
          if (length(dn <- dim(x)) < 2)
              stop("`x' must be an external array of at least two dimensions")
          if (dims < 1 || dims > length(dn) - 1) 
              stop("invalid `dims'")
          p <- prod(dn[-(1:dims)])
          dn <- dn[1:dims]
          x <- .Call("rowSums", x, getNativeStorageMethods(x),
                     prod(dn), p, na.rm,
                     PACKAGE="externalVector")
          if (length(dn) > 1) {
              dim(x) <- dn
              dimnames(x) <- dimnames(x)[1:dims]
          } else names(x) <- dimnames(x)[[1]]
          x
      })

if (!isGeneric("rowMeans"))
    setGeneric("rowMeans",
               function(x, na.rm, dims)
               standardGeneric("rowMeans"))

setMethod("rowMeans",
          signature(x="externalVectorWithStorage",
                    na.rm="missing",
                    dims = "missing"),
          function(x, na.rm, dims)
      {
          callGeneric(x, na.rm=FALSE, dims=1)
      })

setMethod("rowMeans",
          signature(x="externalVectorWithStorage",
                    na.rm="missing"),
          function(x, na.rm, dims)
      {
          callGeneric(x, na.rm=FALSE, dims=dims)
      })

setMethod("rowMeans",
          signature(x="externalVectorWithStorage",
                    dims="missing"),
          function(x, na.rm, dims)
      {
          callGeneric(x, na.rm=na.rm, dims=1)
      })

setMethod("rowMeans",
          signature(x="externalVectorWithStorage",
                    na.rm="logical",
                    dims="numeric"),
          function(x, na.rm, dims)
      {
          if (length(dn <- dim(x)) < 2)
              stop("`x' must be an external array of at least two dimensions")
          if (dims < 1 || dims > length(dn) - 1) 
              stop("invalid `dims'")
          p <- prod(dn[-(1:dims)])
          dn <- dn[1:dims]
          x <- .Call("rowMeans", x, getNativeStorageMethods(x),
                     prod(dn), p, na.rm,
                     PACKAGE="externalVector")
          if (length(dn) > 1) {
              dim(x) <- dn
              dimnames(x) <- dimnames(x)[1:dims]
          } else names(x) <- dimnames(x)[[1]]
          x
      })

setMethod("apply",
          signature(X="externalVector"),
          function (X, MARGIN, FUN, ...)  
          callGeneric(as.Rmatrix(X), MARGIN, FUN, ...))

setMethod("log",
          signature(x="externalVectorWithStorage"),
          function (x, base) 
      {
          klass <- class(defaultElement(x))
          if (!(klass %in% c("logical", "integer", "numeric",
                             "complex")))
              stop("Non-numeric argument to mathematical function")
          if (missing(base)) {
              if (length(x) == 0)
                  return(numeric(0))
              ans <- rebind(x,
                            type = if (klass == "complex") {
                                complex(1)
                            } else numeric(1))
              .Call("externalVectorMath", "log",
                    x, getNativeStorageMethods(x@storage),
                    ans, getNativeStorageMethods(ans@storage),
                    PACKAGE="externalVector")
          } else {
              if (length(base) == 0)
                  stop("illegal 2nd arg of length 0")
              if (length(x) == 0)
                  return(numeric(0))
              if (is(base, "externalVectorWithStorage")) {
                  klass <- class(defaultElement(base))
                  if (!(klass %in% c("logical", "integer", "numeric",
                                     "complex")))
                      stop("Non-numeric argument to mathematical function")
                  ans <- .Call('externalVectorLog', x,
                               base,
                               PACKAGE="externalVector")
                  if (is(ans, "externalStorage")) {
                      n <- length(x)
                      x@storage <- ans
                      if (length(base) > n) {
                          dim(x) <- dim(base)
                          dimnames(x) <- dimnames(base)
                          names(x) <- names(base)
                      }
                      x
                  } else ans
              } else {
                  klass <- class(base)
                  if (!(klass %in% c("logical", "integer", "numeric",
                                     "complex")))
                      stop("Non-numeric argument to mathematical function")
                  ans <- .Call('externalVectorLog', x,
                               base,
                               PACKAGE="externalVector")
                  if (is(ans, "externalStorage")) {
                      n <- ans@length
                      if (length(x) == n) {
                          x@storage <- ans
                          x
                      } else {
                          base@storage <- ans
                          base
                      }
                  } else {
                      ans
                  }
              }
          }
      })

setMethod("lapply",
          signature(X="externalVector"),
          function (X, FUN, ...) 
      {
          X <- force(X)
          callNextMethod()
      })

setMethod("lapply",
          signature(X="externalVectorWithStorage"),
          function (X, FUN, ...) 
      {
          FUN <- match.fun(FUN)
          mcall <- match.call()
          mcall$FUN <- NULL
          mcall[[1]] <- FUN
          names(mcall)[which(names(mcall) == "X")] <- ""
          rval <- .Call("externalVector_lapply",
                        X, length(X), mcall, parent.frame(),
                        PACKAGE="externalVector")
          names(rval) <- names(X)
          return(rval)
      })

setMethod("sapply",
          signature(X="externalVector"),
          function (X, FUN, ..., simplify, USE.NAMES) 
      {
          FUN <- match.fun(FUN)
          answer <- lapply(X, FUN, ...)
          if (USE.NAMES && is.character(defaultElement(X)) &&
              is.null(names(answer))) 
              names(answer) <- as.Rvector(X)
          if (simplify && length(answer) &&
              length(common.len <- unique(unlist(lapply(answer, 
                                                        length)))) == 1) {
              if (common.len == 1) 
                  unlist(answer, recursive = FALSE)
              else if (common.len > 1) 
                  array(unlist(answer, recursive = FALSE),
                        dim = c(common.len, length(X)),
                        dimnames = if (!(is.null(n1 <- names(answer[[1]])) & 
                                         is.null(n2 <- names(answer)))) {
                            list(n1, n2)
                        })
              else answer
          } else answer
      })

setMethod("mean",
          signature(x="externalVector"),
          function (x, trim = 0, na.rm = FALSE, ...) 
          mean(as.Rvector(as.vector(x)), trim, na.rm, ...))

setMethod("median",
          signature(x="externalVector"),
          function (x, na.rm) 
          mean(as.Rvector(as.vector(x)), na.rm))
.First.lib <-
    function(lib, pkg, where)
{
    library.dynam("externalVector", pkg, lib)
}
